use core::panic;
use std::{any::Any, fmt::Display};

use crate::{
    typecheck::{self, Checkable, TypeStack},
    Symtab,
};

#[derive(PartialEq, Eq, Debug, Clone)]
// enum values are parser values, not compiler values,
// e.g. for `Str(String)` the `String` value will be put somewhere in memory
// and a pointer to it will be put on the stack
pub enum Value {
    Nat(u16), // 16-bit natural number
    Int(i16), // 16-bit twos-complement integer
    Bool(bool),
    Op(String),
    Char(char), // 8-bit georgescii character padded with leading zeros (might change later)
    Str(String), // 16-bit pointer to a string
    Word(String), // 16-bit pointer to a word
}

#[derive(PartialEq, Eq, Debug, Clone, PartialOrd, Ord)]
pub enum VType {
    Nat,
    Int,
    Bool,
    Char,
    Str,
}

#[derive(Eq, PartialEq, Debug, Clone)]
pub struct WType {
    pub pop: Vec<VType>,
    pub push: Vec<VType>,
}

impl WType {
    pub fn new() -> Self {
        WType {
            pop: vec![],
            push: vec![],
        }
    }

    // Adds a `push` type
    pub fn push(mut self, mut t: Vec<VType>) -> Self {
        self.push.append(&mut t);
        self
    }

    // Adds a `pop` type (Note: does not actually pop anything)
    pub fn pop(mut self, mut t: Vec<VType>) -> Self {
        self.pop.append(&mut t);
        self
    }
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub enum Effect {
    Paint,
    Sing,
    Store,
    Do,
}

#[derive(Debug, Eq, PartialEq, Clone)]
pub struct WordDef {
    pub name: String,
    pub values: Vec<Value>,
    pub r#type: WType,
    pub effects: Vec<Effect>,
}

impl WordDef {
    pub fn new<S>(name: S, values: Vec<Value>, r#type: WType, effects: Vec<Effect>) -> Self
    where
        S: Into<String>,
    {
        WordDef {
            name: name.into(),
            values,
            r#type,
            effects,
        }
    }

    fn flatten_values(&self, symtab: &Symtab) -> Vec<Value> {
        let mut vals = vec![];
        for value in self.values.iter() {
            if let Value::Word(string) = value {
                let symbol = symtab.get(string);
                let mut child_vals = symbol.flatten_values(symtab);
                vals.append(&mut child_vals);
            } else {
                vals.push(value.clone());
            }
        }
        vals
    }

    pub fn flatten(&self, symtab: &Symtab) -> WordDef {
        let values = self.flatten_values(symtab);
        WordDef::new(
            self.name.clone(),
            values,
            self.r#type.clone(),
            self.effects.clone(),
        )
    }
}

impl Display for WordDef {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        writeln!(f, "{:#?}", self)?;
        Ok(())
    }
}

impl Checkable<VType> for WordDef {
    fn check(&mut self, mut stack: TypeStack<VType>) -> Result<TypeStack<VType>, String> {
        for value in self.values.iter() {
            println!(
                "WORDDEF: checking value {:?} in word {:?}, current stack is {:?}",
                value, self.name, stack
            );
            if let Value::Word(_) = value {
                panic!("Don't typecheck on an unflattened word!");
            }
            match value {
                Value::Op(op) => {
                    if stack.len() < 2 {
                        return Err(format!(
                            "Checking def {:?}, stack is {:?}, expected a stack with 2 elements, got only {:?}",
                            self.name,
                            stack,
                            stack.len()
                        ));
                    } else {
                        match op.as_str() {
                            "+" | "*" => {
                                println!("WORDDEF: checking Op");
                                stack = stack.test_consume(
                                    TypeStack::new().push(VType::Nat).push(VType::Nat),
                                )?;
                                stack = stack.push(VType::Nat);
                            }
                            "&" | "||" => {
                                stack = stack.test_consume(
                                    TypeStack::new().push(VType::Bool).push(VType::Bool),
                                )?;
                                stack = stack.pop().pop().push(VType::Bool);
                            }
                            _ => return Err(format!("unknown opcode {:?}", op)),
                        }
                    }
                }
                &Value::Nat(_) => stack = stack.push(VType::Nat),
                &Value::Int(_) => stack = stack.push(VType::Int),
                &Value::Bool(_) => stack = stack.push(VType::Int),
                &Value::Str(_) => stack = stack.push(VType::Str),
                &Value::Char(_) => stack = stack.push(VType::Char),
                &Value::Word(_) => unreachable!(),
            };
        }
        Ok(stack)
    }
}
