use std::collections::HashMap;

// TODO:
//  - add error types and error handling
//  - figure out if we can use a single ast or should make a second ast for compilation, then
//  convert to that after parsing/do another parsing step but on the ast made the first time
//  (i also don't really know what i'm doing so maybe u never do that or always do that or
//  something i don't understand yet)
//  - improve whitespace parsing
//  - think more about language rules:
//      - what types actually need to be exposed to the "user" (me) and what types can be internal
//      to the compiler (e.g. char vs str, where a str of length 1 could be treated as a char
//      internally (put on the stack as a value instead of put somewhere in memory))

use chumsky::{
    prelude::*,
    text::{ident, keyword},
};

use crate::{
    ast::*,
    typecheck::{Checkable, TypeStack},
};

#[derive(PartialEq, Debug, Clone)]
pub struct Symtab(HashMap<String, WordDef>);

impl Symtab {
    fn new() -> Self {
        Symtab(HashMap::new())
    }

    pub fn get(&self, string: &String) -> &WordDef {
        self.0.get(string).unwrap()
    }

    pub fn add_def(&mut self, def: WordDef) {
        let key = def.name.clone();
        self.0.insert(key, def);
    }

    pub fn flatten_refs(&mut self) {
        let mut new_symtab = Symtab::new();
        for (_string, def) in self.0.iter() {
            new_symtab.add_def(def.flatten(self));
        }
        // this is an abomination, there must be a better way
        self.0.clear();
        self.0.extend(new_symtab.0);
    }
}

impl From<Vec<WordDef>> for Symtab {
    fn from(value: Vec<WordDef>) -> Self {
        let symtab: HashMap<String, WordDef> = value
            .iter()
            .map(|x| (x.name.to_owned(), x.to_owned()))
            .collect();
        Symtab(symtab)
    }
}

#[derive(Debug, PartialEq)]
pub struct Program {
    symtab: Symtab,
    body: Vec<Value>,
}

impl Program {
    fn new(defs: Vec<WordDef>, body: Vec<Value>) -> Self {
        let symtab = Symtab::from(defs);
        Program { symtab, body }
    }

    fn reduce_body(&mut self) {
        let mut vals = vec![];
        for value in self.body.iter() {
            if let Value::Word(string) = value {
                let symbol = self.symtab.get(string);
                let mut child_vals = symbol.flatten(&self.symtab).values;
                vals.append(&mut child_vals);
            } else {
                vals.push(value.clone());
            }
        }
        self.body = vals;
    }
}

impl<'a> Checkable<VType> for Program {
    fn check(
        &mut self,
        mut stack: crate::typecheck::TypeStack<VType>,
    ) -> Result<crate::typecheck::TypeStack<VType>, String> {
        //TODO: https://trykv.medium.com/algorithms-on-graphs-directed-graphs-and-cycle-detection-3982dfbd11f5
        fn cyclic_graph_check(symtab: &Symtab) -> Result<(), String> {
            let mut visited: Vec<&WordDef> = vec![];
            let mut rec_stack: Vec<&WordDef> = vec![];
            for (_, def) in symtab.0.iter() {
                if !visited.contains(&def) {
                    dfs_cycle_check(def, &mut visited, &mut rec_stack, symtab)?;
                }
            }
            Ok(())
        }
        fn dfs_cycle_check<'a>(
            def: &'a WordDef,
            visited: &mut Vec<&'a WordDef>,
            rec_stack: &mut Vec<&'a WordDef>,
            symtab: &'a Symtab,
        ) -> Result<(), String> {
            visited.push(def);
            rec_stack.push(def);

            for val in def.values.iter() {
                if let Value::Word(name) = val {
                    let next_def = symtab.get(name);
                    if !visited.contains(&next_def) {
                        dfs_cycle_check(next_def, visited, rec_stack, symtab)?;
                    } else if rec_stack.contains(&next_def) {
                        return Err(format!(
                            "illegal recursion detected! definitions {}create a reference cycle",
                            rec_stack
                                .iter()
                                .map(|def| {
                                    let mut name = def.name.clone();
                                    name.insert(0, '"');
                                    name.push_str("\" ");
                                    name
                                })
                                .collect::<String>()
                        ));
                    }
                }
            }

            rec_stack.pop();
            Ok(())
        }

        cyclic_graph_check(&self.symtab)?;

        self.symtab.flatten_refs();

        println!(
            "we have flattened refs, here's the symtab: {:#?}\n",
            self.symtab
        );

        // then check that all symtab defs are sound
        // at this point they shouldn't have any references,
        // and if they do we will panic (see the Checkable impl for WordDef)
        for (name, def) in self.symtab.0.iter_mut() {
            let local_stack: TypeStack<VType> = def.r#type.pop.clone().into();
            println!(
                "PARSED: checking {:?}\ncurrent stack: {local_stack:?}\nword: {:?}",
                name, def
            );
            let result_stack = def.check(local_stack)?;
            if let Err(error) = result_stack.test(&def.r#type.push.clone().into()) {
                println!("{error:?}");
                return Err(error);
            }
        }

        self.reduce_body();

        // then we'll check that the body is sound with the given stack
        // maybe in the future i'll change this trait so there isn't a stack
        // param and the implementer picks what stack to check against
        //
        // TODO: this block also is shared behavior between basically all checkables but potentially with
        // different internal types for T, will have to figure out how to dedup this later
        for value in self.body.iter() {
            match value {
                Value::Op(op) => {
                    if stack.len() < 2 {
                        return Err(format!(
                            "expected a stack with 2 elements, got only {:?}",
                            stack.len()
                        ));
                    } else {
                        match op.as_str() {
                            "+" | "*" => {
                                stack = stack.test_consume(
                                    TypeStack::new().push(VType::Nat).push(VType::Nat),
                                )?;
                                stack = stack.pop().pop().push(VType::Nat);
                            }
                            "&" | "||" => {
                                stack = stack.test_consume(
                                    TypeStack::new().push(VType::Bool).push(VType::Bool),
                                )?;
                                stack = stack.pop().pop().push(VType::Bool);
                            }
                            _ => return Err(format!("unknown opcode {:?}", op)),
                        }
                    }
                }
                &Value::Nat(_) => stack = stack.push(VType::Nat),
                &Value::Int(_) => stack = stack.push(VType::Int),
                &Value::Bool(_) => stack = stack.push(VType::Bool),
                &Value::Str(_) => stack = stack.push(VType::Str),
                &Value::Char(_) => stack = stack.push(VType::Char),
                &Value::Word(_) => unreachable!(),
            };
        }
        Ok(stack)
    }
}

pub fn parse<S>(input: S) -> Result<Program, Vec<Simple<char>>>
where
    S: ToString,
{
    let parsed = match parser().parse(input.to_string()) {
        Ok(parsed) => parsed,
        Err(error) => return Err(error),
    };

    Ok(parsed)
}

pub fn parser() -> impl Parser<char, Program, Error = Simple<char>> {
    let name = ident().labelled("word_name");
    let value = {
        // nats will be coerced to ints at compile time depending on word type
        let nat = text::int(10).map(|s: String| Value::Nat(s.parse().unwrap()));

        // vice versa for non-negative ints
        let int = just("-").ignore_then(
            text::int::<char, Simple<char>>(10).map(|s: String| Value::Int(s.parse().unwrap())),
        );

        let op = one_of::<char, &str, Simple<char>>("*+-/&|<>").map(|s| Value::Op(s.to_string()));

        let str_or_char = just::<char, char, Simple<char>>('"')
            .ignore_then(none_of('"').repeated())
            .then_ignore(just('"'))
            .map(|s: Vec<char>| match s.len() {
                1 => Value::Char(s[0]),
                _ => Value::Str(s.into_iter().collect::<String>()),
            });

        let word = name.map(|n: String| Value::Word(n));

        let bool = keyword::<_, _, Simple<char>>("true")
            .map(|_| Value::Bool(true))
            .or(keyword("false").map(|_| Value::Bool(false)));

        nat.or(int).or(bool).or(str_or_char).or(word).or(op)
    };
    let value_seperator = text::newline()
        .repeated()
        .at_least(2)
        .not()
        .rewind()
        .then_ignore(
            // TODO: figure out if this could be simplified
            choice((
                just(" ")
                    .repeated()
                    .then_ignore(just("\n").repeated().exactly(1).or_not()),
                just("\n")
                    .repeated()
                    .exactly(1)
                    .then_ignore(just(" ").repeated().or_not()),
            ))
            .then_ignore(just(" ").repeated()),
        );

    let body = value_seperator
        .or_not()
        .ignored()
        .then(value)
        .map(|(_, v)| v)
        .repeated()
        .then_ignore(
            just(" ")
                .repeated()
                .ignored()
                .then(text::newline().repeated().at_least(2).or_not()),
        );
    let word_def = {
        let pop_types = {
            let pop_type = keyword("nat")
                .to(VType::Nat)
                .or(keyword("int").to(VType::Int))
                .or(keyword("bool").to(VType::Bool))
                .or(keyword("char").to(VType::Char))
                .or(keyword("str").to(VType::Str));

            pop_type
                .padded()
                .repeated()
                .collect::<Vec<VType>>()
                .labelled("pop_types")
                .boxed()
        };

        let push_types = {
            let push_type = keyword("nat")
                .to(VType::Nat)
                .or(keyword("int").to(VType::Int))
                .or(keyword("char").to(VType::Char))
                .or(keyword("str").to(VType::Str));
            push_type
                .padded()
                .repeated()
                .collect::<Vec<VType>>()
                .labelled("push_types")
                .boxed()
        };

        let effects = {
            let effect_keyword = keyword("paint")
                .to(Effect::Paint)
                .or(keyword("sing").to(Effect::Sing))
                .or(keyword("store").to(Effect::Store))
                .or(keyword("do").to(Effect::Do));

            let effect = just("~").ignore_then(effect_keyword).labelled("effect");

            effect.padded().repeated().labelled("effects").boxed()
        };

        let definition = text::whitespace()
            .ignore_then(name)
            .then_ignore(just(" "))
            .then(pop_types)
            .then_ignore(keyword("is").or(keyword("are")).padded())
            .then(push_types)
            .then(effects)
            .then_ignore(just(":"))
            .map(|(((name, pop_types), push_types), effects)| {
                (name, pop_types, push_types, effects)
            });

        definition
            .then(body.clone())
            .map(|((name, pop_types, push_types, effects), body)| {
                WordDef::new(
                    name,
                    body,
                    WType::new().push(push_types).pop(pop_types),
                    effects,
                )
            })
    };
    word_def
        .repeated()
        .then(body)
        .map(|(defs, body): (Vec<WordDef>, Vec<Value>)| Program::new(defs, body))
}

#[cfg(test)]
mod tests {
    use crate::typecheck::TypeStack;

    use super::*;

    #[test]
    fn test_parser() {
        let input = "
                    a is nat: 5 7 * 

                    b is nat: 
                        5 a * 

                    a
                    ";

        let ast = vec![
            WordDef::new(
                "a",
                vec![Value::Nat(5), Value::Nat(7), Value::Op("*".to_string())],
                WType::new().push(vec![VType::Nat]),
                vec![],
            ),
            WordDef::new(
                "b",
                vec![
                    Value::Nat(5),
                    Value::Word("a".to_string()),
                    Value::Op("*".to_string()),
                ],
                WType::new().push(vec![VType::Nat]),
                vec![],
            ),
        ];
        let body: Vec<Value> = vec![Value::Word("a".to_string())];
        println!("sound: {:?}\n", parser().parse(input).unwrap());
        assert_eq!(parser().parse(input).unwrap(), Program::new(ast, body));
    }

    #[test]
    fn test_typecheck() {
        let sound = "
                    a is nat: 5 7 * 

                    b nat nat is nat: 
                        a * 

                    a 5 * 
                    ";

        let unsound_defs = "
                    a is nat nat: 5 7 *

                    b nat is nat:
                        a *

                    a 5 *
                    ";

        let unsound_body = "
                    a is nat: 5 7 *

                    b nat is nat:
                        a *

                    a 5 * *
                    ";

        let unsound_body_and_defs = "
                    a is nat nat: 5 7 *

                    b nat is nat:
                        a *

                    a 5 * *
                    ";

        fn typecheck(input: &str, sound: bool) {
            let mut parsed = parse(input).unwrap();
            parsed.symtab.flatten_refs();
            parsed.reduce_body();

            let stack = TypeStack::new();
            if sound {
                assert!(parsed.check(stack).is_ok());
            } else {
                assert!(parsed.check(stack).is_err());
            }
        }

        typecheck(sound, true);
        typecheck(unsound_defs, false);
        typecheck(unsound_body, false);
        typecheck(unsound_body_and_defs, false);
    }

    #[test]
    fn test_illegal_recursion() {
        let illegal = "
                    a is: b

                    b is: a

                    a
                    ";

        let illegal_multilevel = "
                    a is: b

                    b is: c

                    c is: a

                    a
                    ";

        fn typecheck(input: &str) {
            let mut parsed = parse(input).unwrap();
            let stack = TypeStack::new();
            println!("{:?}", parsed.check(TypeStack::new()));
            assert!(parsed.check(stack).is_err());
        }

        typecheck(illegal);
        typecheck(illegal_multilevel);
    }
}
